/* Copyright (c) 2005 PrimeBase Technologies GmbH
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2008-01-24	Paul McCullagh
 *
 * Row lock functions.
 *
 * H&G2JCtL
 */
#ifndef __xt_lock_h__
#define __xt_lock_h__

#include "xt_defs.h"
#include "util_xt.h"

struct XTThread;
struct XTDatabase;
struct XTOpenTable;
struct XTXactData;

/*
 * [(9)]
 *
 * These are perminent row locks. They are set on rows for 2 reasons:
 *
 * 1. To lock a row that is being updated. The row is locked
 *    when it is read, until the point that it is updated. If the row
 *    is not updated, the lock is removed.
 *    This prevents an update coming between which will cause an error
 *    on the first thread.
 *
 * 2. The locks are used to implement SELECT FOR UPDATE.
 */

/*
 * A lock that is set in order to perform an update is a temporary lock.
 * This lock will be removed once the update of the record is done.
 * The objective is to prevent some other thread from changine the
 * record between the time the record is read and updated. This is to
 * prevent unncessary "Record was updated" errors.
 *
 * A permanent lock is set by a SELECT FOR UPDATE. These locks are
 * held until the end of the transaction.
 *
 * However, a SELECT FOR UPDATE will pop its lock stack before
 * waiting for a transaction that has updated a record.
 * This is to prevent the deadlock that can occur because a
 * SELECT FOR UPDATE locks groups of records (I mean in general the
 * locks used are group locks).
 *
 * This means a SELECT FOR UPDATE can get ahead of an UPDATE as far as
 * locking is concerned. Example:
 *
 * Record 1,2 and 3 are in group A.
 *
 * T1: UPDATES record 2.
 * T2: SELECT FOR UPDATE record 1, which locks group A.
 * T2: SELECT FOR UPDATE record 2, which must wait for T1.
 * T1: UPDATES record 3, which musts wait because of group lock A.
 *
 * To avoid deadlock, T2 releases its group lock A before waiting for
 * record 2. It then regains the lock after waiting for record 2.
 *
 * (NOTE: Locks are no longer released. Please check this comment:
 * {RELEASING-LOCKS} in lock_xt.cc. )
 *
 * However, release group A lock mean first releasing all locks gained
 * after group a lock.
 *
 * For example: a thread locks groups: A, B and C. To release group B
 * lock the thread must release C as well. Afterwards, it must gain
 * B and C again, in that order. This is to ensure that the lock
 * order is NOT changed!
 *
 */
#define XT_NO_LOCK					0
#define XT_TEMP_LOCK				1								/* A temporary lock */
#define XT_PERM_LOCK				2								/* A permanent lock */

typedef struct XTPermRowLock {
	xtTableID				pr_tab_id;
	xtWord4					pr_group;
} XTPermRowLockRec, *XTPermRowLockPtr;

typedef struct XTRowLockList : public XTBasicList {
	u_int					rll_release_point;						/* The point to which locks have been released. */

	void	xt_remove_all_locks(struct XTDatabase *db);
} XTRowLockListRec, *XTRowLockListPtr;

xtBool			xt_init_row_lock_list(XTRowLockListPtr rl);
void			xt_exit_row_lock_list(XTRowLockListPtr rl);

typedef struct XTRowLocks {
	xtWord1					tab_lock_perm[XT_ROW_LOCK_COUNT];		/* Byte set to 1 for permanent locks. */
	struct XTXactData		*tab_row_locks[XT_ROW_LOCK_COUNT];		/* The transactions that have locked the specific rows. */

	int		xt_set_temp_lock(struct XTOpenTable *ot, xtRowID row, xtXactID *xn_id, XTRowLockListPtr lock_list);
	void	xt_remove_temp_lock(struct XTOpenTable *ot);
	xtBool	xt_make_lock_permanent(struct XTOpenTable *ot, XTRowLockListPtr lock_list);
	int		xt_is_locked(struct XTOpenTable *ot, xtRowID row, xtXactID *xn_id);
	int		xt_release_locks(struct XTOpenTable *ot, xtRowID row, XTRowLockListPtr lock_list);
	xtBool	xt_regain_locks(struct XTOpenTable *ot, int *lock_type, xtXactID *xn_id, XTRowLockListPtr lock_list);
} XTRowLocksRec, *XTRowLocksPtr;

xtBool			xt_init_row_locks(XTRowLocksPtr rl);
void			xt_exit_row_locks(XTRowLocksPtr rl);

#define XT_NO_LOCK				0
#define XT_WANT_LOCK			1
#define XT_HAVE_LOCK			2
#define XT_WAITING				3

/* Possibilities are 2 = align 4 or 2 = align 8 */
#define XT_XS_LOCK_SHIFT		2
#define XT_XS_LOCK_ALIGN		(1 << XT_XS_LOCK_SHIFT)

typedef struct XTRWMutex {
#ifdef DEBUG
	struct XTThread				*xs_lock_thread;
	u_int						xs_inited;
#endif
	xt_mutex_type				xs_lock;
	xt_cond_type				xs_cond;
	volatile xtWord4			xs_state;
	volatile xtThreadID			xs_xlocker;
	union {
#if XT_XS_LOCK_ALIGN == 4
		volatile xtWord4		*xs_rlock_align;
#else
		volatile  xtWord8		*xs_rlock_align;
#endif
		volatile  xtWord1		*xs_rlock;
	}							x;
} XTRWMutexRec, *XTRWMutexPtr;

void xt_rwmutex_init(struct XTThread *self, XTRWMutexPtr xsl);
void xt_rwmutex_free(struct XTThread *self, XTRWMutexPtr xsl);
xtBool xt_rwmutex_xlock(XTRWMutexPtr xsl, xtThreadID thd_id);
xtBool xt_rwmutex_slock(XTRWMutexPtr xsl, xtThreadID thd_id);
xtBool xt_rwmutex_unlock(XTRWMutexPtr xsl, xtThreadID thd_id);

#ifdef XT_WIN
#define XT_SPL_WIN32_ASM
#else
#if defined(__GNUC__) && (defined(__x86_64__) || defined(__i386__))
#define XT_SPL_GNUC_X86
#else
#define XT_SPL_DEFAULT
#endif
#endif

typedef struct XTSpinLock {
	volatile xtWord4			spl_lock;
#ifdef XT_SPL_DEFAULT
	xt_mutex_type				spl_mutex;
#endif
#ifdef DEBUG
	struct XTThread				*spl_locker;
#endif
} XTSpinLockRec, *XTSpinLockPtr;

void	xt_spinlock_init(struct XTThread *self, XTSpinLockPtr sp);
void	xt_spinlock_free(struct XTThread *self, XTSpinLockPtr sp);
xtBool	xt_spinlock_spin(XTSpinLockPtr spl);
#ifdef DEBUG
void	xt_spinlock_set_thread(XTSpinLockPtr spl);
#endif

/*
 * This macro is to remind me where it was safe
 * to use a read lock!
 */
#define xt_lck_slock		xt_spinlock_lock

/* We must make sure that we write one byte!
 * If the compiler loads 4 bytes, and then writes,
 * bytes can be overwritten.
 */
inline void xt_write1(volatile xtWord1 *mptr, xtWord1 val)
{
#ifdef XT_SPL_WIN32_ASM
	__asm MOV  ECX, mptr
	__asm MOV  DL, val
	__asm MOV  BYTE PTR [ECX], DL
#elif defined(XT_SPL_GNUC_X86)
	asm volatile ("movb %0,%1" : : "m" (*mptr), "r" (val) : "memory");
#else
	*mptr = val;
#endif
}

inline void xt_atomic_set1(volatile xtWord1 *mptr, xtWord1 val)
{
#ifdef XT_SPL_WIN32_ASM
	__asm MOV  ECX, mptr
	__asm MOV  DL, val
	__asm XCHG DL, BYTE PTR [ECX]
#elif defined(XT_SPL_GNUC_X86)
	asm volatile ("xchgb %1,%0" : "=r" (val) : "m" (*mptr), "0" (val) : "memory");
#else
	*mptr = val;
#endif
}

inline xtWord1 xt_atomic_get1(volatile xtWord1 *mptr)
{
	xtWord1 val;

#ifdef XT_SPL_WIN32_ASM
	__asm MOV ECX, mptr
	__asm MOV DL, BYTE PTR [ECX]
	__asm MOV val, DL
#elif defined(XT_SPL_GNUC_X86)
	asm volatile ("movb %1,%0" : "=r" (val) : "m" (*mptr) : "memory");
#else
	val = *mptr;
#endif
	return val;
}

inline void xt_atomic_set4(volatile xtWord4 *mptr, xtWord4 val)
{
#ifdef XT_SPL_WIN32_ASM
	__asm MOV  ECX, mptr
	__asm MOV  EDX, val
	__asm XCHG EDX, DWORD PTR [ECX]
	//__asm MOV  DWORD PTR [ECX], EDX
#elif defined(XT_SPL_GNUC_X86)
	asm volatile ("xchgl %1,%0" : "=r" (val) : "m" (*mptr), "0" (val) : "memory");
	//asm volatile ("movl %0,%1" : "=r" (val) : "m" (*mptr) : "memory");
#else
	*mptr = val;
#endif
}

inline xtWord4 xt_atomic_get4(volatile xtWord4 *mptr)
{
	xtWord4 val;

#ifdef XT_SPL_WIN32_ASM
	__asm MOV ECX, mptr
	__asm MOV EDX, DWORD PTR [ECX]
	__asm MOV val, EDX
#elif defined(XT_SPL_GNUC_X86)
	asm volatile ("movl %1,%0" : "=r" (val) : "m" (*mptr) : "memory");
#else
	val = *mptr;
#endif
	return val;
}

/* Code for test and set is derived from code by Larry Zhou and
 * Google: http://code.google.com/p/google-perftools
 */
inline xtWord4 xt_spinlock_set(XTSpinLockPtr spl)
{
	xtWord4				prv;
	volatile xtWord4	*lck;
				
	lck = &spl->spl_lock;
#ifdef XT_SPL_WIN32_ASM
	__asm MOV  ECX, lck
	__asm MOV  EDX, 1
	__asm XCHG EDX, DWORD PTR [ECX]
	__asm MOV  prv, EDX
#elif defined(XT_SPL_GNUC_X86)
	prv = 1;
	asm volatile ("xchgl %1,%0" : "=r" (prv) : "m" (*lck), "0" (prv) : "memory");
#else
	/* The default implementation just uses a mutex, and
	 * does not spin! */
	xt_lock_mutex_ns(&spl->spl_mutex);
	/* We have the lock */
	*lck = 1;
	prv = 0;
#endif
#ifdef DEBUG
	xt_spinlock_set_thread(spl);
#endif
	return prv;
}

inline xtWord4 xt_spinlock_reset(XTSpinLockPtr spl)
{
	xtWord4				prv;
	volatile xtWord4	*lck;
				
#ifdef DEBUG
	spl->spl_locker = NULL;
#endif
	lck = &spl->spl_lock;
#ifdef XT_SPL_WIN32_ASM
	__asm MOV  ECX, lck
	__asm MOV  EDX, 0
	__asm XCHG EDX, DWORD PTR [ECX]
	__asm MOV  prv, EDX
#elif defined(XT_SPL_GNUC_X86)
	prv = 0;
	asm volatile ("xchgl %1,%0" : "=r" (prv) : "m" (*lck), "0" (prv) : "memory");
#else
	*lck = 0;
	xt_unlock_mutex_ns(&spl->spl_mutex);
	prv = 1;
#endif
	return prv;
}

/*
 * Return FALSE, and register an error on failure.
 */
inline xtBool xt_spinlock_lock(XTSpinLockPtr spl)
{
	if (!xt_spinlock_set(spl))
		return OK;
	return xt_spinlock_spin(spl);
}

inline void xt_spinlock_unlock(XTSpinLockPtr spl)
{
	xt_spinlock_reset(spl);
}

void xt_unit_test_read_write_locks(struct XTThread *self);
void xt_unit_test_mutex_locks(struct XTThread *self);
void xt_unit_test_create_threads(struct XTThread *self);

#define XT_FAST_LOCK_MAX_WAIT	100

typedef struct XTFastLock {
	XTSpinLockRec				fal_spinlock;
	struct XTThread				*fal_locker;

	XTSpinLockRec				fal_wait_lock;
	u_int						fal_wait_count;
	u_int						fal_wait_wakeup;
	u_int						fal_wait_alloc;
	struct XTThread				*fal_wait_list[XT_FAST_LOCK_MAX_WAIT];
} XTFastLockRec, *XTFastLockPtr;

void	xt_fastlock_init(struct XTThread *self, XTFastLockPtr spl);
void	xt_fastlock_free(struct XTThread *self, XTFastLockPtr spl);
void	xt_fastlock_wakeup(XTFastLockPtr spl);
xtBool	xt_fastlock_spin(XTFastLockPtr spl, struct XTThread *thread);

inline xtBool xt_fastlock_lock(XTFastLockPtr fal, struct XTThread *thread)
{
	if (!xt_spinlock_set(&fal->fal_spinlock)) {
		fal->fal_locker = thread;
		return OK;
	}
	return xt_fastlock_spin(fal, thread);
}

inline void xt_fastlock_unlock(XTFastLockPtr fal, struct XTThread *thread __attribute__((unused)))
{
	if (fal->fal_wait_count)
		xt_fastlock_wakeup(fal);
	else {
		fal->fal_locker = NULL;
		xt_spinlock_reset(&fal->fal_spinlock);
	}
}

typedef struct XTSpinRWLock {
	XTSpinLockRec				srw_lock;
	volatile xtThreadID			srw_xlocker;
	XTSpinLockRec				srw_state_lock;
	volatile u_int				srw_state;
	union {
#if XT_XS_LOCK_ALIGN == 4
		volatile xtWord4		*srw_rlock_align;
#else
		volatile  xtWord8		*srw_rlock_align;
#endif
		volatile  xtWord1		*srw_rlock;
	} x;
} XTSpinRWLockRec, *XTSpinRWLockPtr;

void xt_spinrwlock_init(struct XTThread *self, XTSpinRWLockPtr xsl);
void xt_spinrwlock_free(struct XTThread *self, XTSpinRWLockPtr xsl);
xtBool xt_spinrwlock_xlock(XTSpinRWLockPtr xsl, xtThreadID thd_id);
xtBool xt_spinrwlock_slock(XTSpinRWLockPtr xsl, xtThreadID thd_id);
xtBool xt_spinrwlock_unlock(XTSpinRWLockPtr xsl, xtThreadID thd_id);

typedef struct XTFastRWLock {
	XTFastLockRec				frw_lock;
	struct XTThread				*frw_xlocker;
	XTSpinLockRec				frw_state_lock;
	volatile u_int				frw_state;
	u_int						frw_read_waiters;
	union {
#if XT_XS_LOCK_ALIGN == 4
		volatile xtWord4		*frw_rlock_align;
#else
		volatile  xtWord8		*frw_rlock_align;
#endif
		volatile  xtWord1		*frw_rlock;
	} x;
} XTFastRWLockRec, *XTFastRWLockPtr;

void xt_fastrwlock_init(struct XTThread *self, XTFastRWLockPtr frw);
void xt_fastrwlock_free(struct XTThread *self, XTFastRWLockPtr frw);
xtBool xt_fastrwlock_xlock(XTFastRWLockPtr frw, struct XTThread *thread);
xtBool xt_fastrwlock_slock(XTFastRWLockPtr frw, struct XTThread *thread);
xtBool xt_fastrwlock_unlock(XTFastRWLockPtr frw, struct XTThread *thread);

#endif
